﻿using System;
using System.Linq;
using System.Windows.Forms;
using System.Collections.ObjectModel;
using System.Xml.Linq;
using System.Xml;
using System.Collections;
using System.ComponentModel;
/*                                  Decision tree                                */
/*                                                                               */
/*                       www.mosismath.com                                       */

namespace WindowsFormsApplication1
{ 
    public partial class Mainwin : Form
    {
        const int MAXRECURSIONS = 9;
        const int USELESSLEAVE = 100;
        int samples = 29;
        int dataLength = 0;
        const int features = 4;
        int minClass = 100000;
        int maxClass = -100000;
        int nbOfClasses;
        int nbOfRecs = 0;
        int maxRecs = 0;
        TData data = new TData();
        BackgroundWorker calcThread;
        TBranch trunc = new TBranch();


        public Mainwin()
        {
            int i,j;
            DataGridViewCell cell;

            InitializeComponent();
            calcThread = new BackgroundWorker();
            calcThread.WorkerReportsProgress = true;
            calcThread.DoWork += worker_Calc;
            calcThread.RunWorkerCompleted += worker_CalculationCompleted;

            data.readData(Application.StartupPath + "\\data.csv");          
            samples = data.values.Count;
            dataLength = data.dataLength;
            dgDataView.RowCount = samples;
            dgDataView.ColumnCount = features + 2;

            dgDataView.Columns[0].HeaderText = "x1";
            dgDataView.Columns[1].HeaderText = "x2";
            dgDataView.Columns[2].HeaderText = "x3";
            dgDataView.Columns[3].HeaderText = "x4";
            dgDataView.Columns[4].HeaderText = "type";
            dgDataView.Columns[5].HeaderText = "output";

            for (i = 0; i < samples; i++)
            {
                dgDataView.Rows[i].HeaderCell.Value = (i + 1).ToString();
                for (j = 0; j <= features; j++)
                {
                    cell = dgDataView[j, i];
                    cell.Value = Math.Round(data.values.ElementAt(i)[j], 4).ToString();
                }
                if (data.values.ElementAt(i)[features] < minClass)
                    minClass = (int)(data.values.ElementAt(i)[features]);
                if (data.values.ElementAt(i)[features] > maxClass)
                    maxClass = (int)(data.values.ElementAt(i)[features]);
            }
            nbOfClasses = maxClass - minClass + 1;
        }

        private void writeSetup(string FileName, TBranch branch)
        {
            XDocument doc = new XDocument();
            doc.Add(new XElement("Treesetting", null));
            using (XmlWriter writer = XmlWriter.Create(FileName))
            {
                doc.Root.Add(branch.writeBranch(0));
                doc.WriteTo(writer);
            }
        }

        private void readSetup(string FileName, TBranch branch)
        {
            XElement doc = XElement.Load(FileName);
            IEnumerable chields = doc.Elements("Branch");
            
            foreach (XElement el in chields)
            {
                branch.loadBranch(el);
            }
        }

        private void calcGini(int iClass, int iProperty, TBranch branch, Collection<double[]> values)
        {
            bool foundClass = false;
            bool othersEmpty = true;
            double classMin =  1e50;
            double classMax = -1e50;
            double gini;
            int[] iOthersCount = new int[nbOfClasses - 1];
            int[] iOthersClass = new int[nbOfClasses - 1];
            int iActIndex = 0;
            int iClassCount = 0;
            int i, j;
            for (i = 0; i < iOthersClass.Length; i++)
                iOthersClass[i] = -1;          
            for (i = 0; i < values.Count; i++)
            {
                if (values.ElementAt(i)[dataLength - 1] == iClass)
                {
                    foundClass = true;
                    if (values.ElementAt(i)[iProperty] < classMin)
                        classMin = values.ElementAt(i)[iProperty];
                    if (values.ElementAt(i)[iProperty] > classMax)
                        classMax = values.ElementAt(i)[iProperty];
                }
            }
           
            if (foundClass)
            {
                for (i = 0; i < values.Count; i++)
                {
                    if (values.ElementAt(i)[dataLength - 1] != iClass)
                    {
                        if ((values.ElementAt(i)[iProperty] > classMax) || (values.ElementAt(i)[iProperty] < classMin))
                        {
                            foundClass = false;
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    foundClass = true;
                            }
                            if(!foundClass)
                            {
                                iOthersClass[iActIndex] = (int)values.ElementAt(i)[dataLength - 1];
                                iActIndex++;
                            }
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    iOthersCount[j]++;
                            }
                        }
                    }
                    else
                        iClassCount++;
                }
                for (j = 0; j < iOthersClass.Length; j++)
                {
                    if (iOthersCount[j] != 0)
                        othersEmpty = false;
                }
                if ((!othersEmpty) || (iClassCount == values.Count))
                {
                    gini = 1.0 - (iClassCount / values.Count) * (iClassCount / values.Count);
                    for (i = 0; i < iOthersClass.Length; i++)
                        gini = gini - (iOthersCount[i] /values.Count) * (iOthersCount[i] / values.Count);
                    branch.purity = gini;
                    branch.min = classMin;
                    branch.max = classMax;
                    branch.iclass = iClass;
                    branch.iProperty = iProperty;
                }
                else
                    branch.purity = USELESSLEAVE;
            }
            else
                branch.purity = USELESSLEAVE;
        }

        private void calcEntropy(int iClass, int iProperty, TBranch branch, Collection<double[]> values)
        {
            double ln_2 = Math.Log(2);
            bool foundClass = false;
            bool othersEmpty = true;
            double classMin = 1e50;
            double classMax = -1e50;
            double entropy, p;
            int[] iOthersCount = new int[nbOfClasses - 1];
            int[] iOthersClass = new int[nbOfClasses - 1];
            int iActIndex = 0;
            int iClassCount = 0;
            int i, j;
            for (i = 0; i < iOthersClass.Length; i++)
                iOthersClass[i] = -1;
            for (i = 0; i < values.Count; i++)
            {
                if (values.ElementAt(i)[dataLength - 1] == iClass)
                {
                    foundClass = true;
                    if (values.ElementAt(i)[iProperty] < classMin)
                        classMin = values.ElementAt(i)[iProperty];
                    if (values.ElementAt(i)[iProperty] > classMax)
                        classMax = values.ElementAt(i)[iProperty];
                }
            }

            if (foundClass)
            {
                for (i = 0; i < values.Count; i++)
                {
                    if (values.ElementAt(i)[dataLength - 1] != iClass)
                    {
                        if ((values.ElementAt(i)[iProperty] > classMax) || (values.ElementAt(i)[iProperty] < classMin))
                        {
                            foundClass = false;
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    foundClass = true;
                            }
                            if (!foundClass)
                            {
                                iOthersClass[iActIndex] = (int)values.ElementAt(i)[dataLength - 1];
                                iActIndex++;
                            }
                            for (j = 0; j < iOthersClass.Length; j++)
                            {
                                if (iOthersClass[j] == (int)values.ElementAt(i)[dataLength - 1])
                                    iOthersCount[j]++;
                            }
                        }
                    }
                    else
                        iClassCount++;
                }
                for (j = 0; j < iOthersClass.Length; j++)
                {
                    if (iOthersCount[j] != 0)
                        othersEmpty = false;
                }
                if ((!othersEmpty) || (iClassCount == values.Count))
                {
                    p = (double)iClassCount / (double)values.Count;
                    entropy = -p * Math.Log(p) / ln_2;
                    for (i = 0; i < iOthersClass.Length; i++)
                    {
                        if (iOthersCount[i] > 0)
                            p = (double)iOthersCount[i] / (double)values.Count;
                        entropy = entropy - p * Math.Log(p) / ln_2;
                    }
                    branch.purity = entropy;
                    branch.min = classMin;
                    branch.max = classMax;
                    branch.iclass = iClass;
                    branch.iProperty = iProperty;
                }
                else
                    branch.purity = 100;
            }
            else
                branch.purity = 100;
        }

        private double BuildBranch(ref TBranch actBranch, Collection<double[]> values)
        {
            int iActClass, iActProperty, j;
            double localGiniL = 0;
            double localGiniR = 0;
            if (nbOfRecs <= MAXRECURSIONS)
            {
                for (iActClass = minClass; iActClass <= maxClass; iActClass++)
                {
                    for (iActProperty = 0; iActProperty < features; iActProperty++)
                    {
                        TBranch branch = new TBranch();
                       // calcGini(iActClass, iActProperty, branch, values);
                        calcEntropy(iActClass, iActProperty, branch, values);
                        if ((branch.purity > 0) && (branch.purity < actBranch.purity))
                        {
                            Collection<double[]> nextValuesL = new Collection<double[]>();
                            Collection<double[]> nextValuesR = new Collection<double[]>();
                            for (j = 0; j < values.Count; j++)
                            {
                                if ((values.ElementAt(j)[branch.iProperty] <= branch.max) && (values.ElementAt(j)[branch.iProperty] >= branch.min))
                                    nextValuesL.Add(values.ElementAt(j));
                                else
                                    nextValuesR.Add(values.ElementAt(j));
                            }
                            nbOfRecs++;
                            if (maxRecs < nbOfRecs)
                                maxRecs = nbOfRecs;
                            if (nextValuesL.Count > 0)
                            {
                                branch.nextBranch[0] = new TBranch();
                                localGiniL = BuildBranch(ref branch.nextBranch[0], nextValuesL) * nextValuesL.Count;
                            }
                            if (nextValuesR.Count > 0)
                            {
                                branch.nextBranch[1] = new TBranch();
                                localGiniR = BuildBranch(ref branch.nextBranch[1], nextValuesR) * nextValuesR.Count;
                            }
                            nbOfRecs--;
                        }
                        branch.purity = branch.purity + (localGiniL + localGiniR) / values.Count;
                        if (actBranch.purity > branch.purity)
                        {
                            actBranch = branch;
                        }
                        if (actBranch.purity == 0)
                        {
                            iActProperty = features;
                            iActClass = maxClass + 1;
                        }
                    }
                }
            }
            else
                actBranch.purity = USELESSLEAVE;
            return actBranch.purity;
        }

        private void worker_Calc(object sender, DoWorkEventArgs e)
        {
            BuildBranch(ref trunc, data.values);
            writeSetup(Application.StartupPath + "\\setup.xml", trunc);
        }

        void worker_CalculationCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            int i, j, flower;
            DataGridViewCell cell;
            for (i = 0; i < data.values.Count; i++)
            {
                flower = trunc.CheckBranch(data.values.ElementAt(i));
                cell = dgDataView[5, i];
                cell.Value = flower.ToString();
            }
            bt_Calc.Enabled = true;
        }

        private void bt_Calc_Click(object sender, EventArgs e)
        {
            bt_Calc.Enabled = false;
            calcThread.RunWorkerAsync();
        }
    }
}
